#include "ShadowMapping.h"
#include "FrameBufferObject.h"
#include "RenderState.h"
#include "RenderTraverser.h"
#include "Light.h"
#include <IL/il.h>
#include <assert.h>

using namespace std;


namespace CHCDemoEngine
{

static GLenum mrt[] = {GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_COLOR_ATTACHMENT2_EXT, GL_COLOR_ATTACHMENT3_EXT};

static void PrintGLerror(char *msg)
{
	GLenum errCode;
	const GLubyte *errStr;
	
	if ((errCode = glGetError()) != GL_NO_ERROR) 
	{
		errStr = gluErrorString(errCode);
		fprintf(stderr,"OpenGL ERROR: %s: %s\n", errStr, msg);
	}
}



static CGprogram sCgShadowProgram;
static CGparameter sShadowParam;


static void GrabDepthBuffer(float *data, GLuint depthTexture)
{
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, depthTexture);

	glGetTexImage(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, GL_FLOAT, data);

	glBindTexture(GL_TEXTURE_2D, 0);
	glDisable(GL_TEXTURE_2D);
}


static void ExportDepthBuffer(float *data, int size)
{
	ilInit();
	assert(ilGetError() == IL_NO_ERROR);

	ILstring filename = ILstring("shadow.tga");
	ilRegisterType(IL_FLOAT);

	const int depth = 1;
	const int bpp = 1;

	if (!ilTexImage(size, size, depth, bpp, IL_LUMINANCE, IL_FLOAT, data))
	{
		cerr << "IL error " << ilGetError() << endl;
	
		ilShutDown();
		assert(ilGetError() == IL_NO_ERROR);

		return;
	}

	if (!ilSaveImage(filename))
	{
		cerr << "TGA write error " << ilGetError() << endl;
	}

	ilShutDown();
	assert(ilGetError() == IL_NO_ERROR);

	cout << "exported depth buffer" << endl;
}



ShadowMapping::ShadowMapping(const AxisAlignedBox3 &sceneBox, int size):
mSceneBox(sceneBox), mSize(size)
{
	mFbo = new FrameBufferObject(size, size, FrameBufferObject::DEPTH_32, true);
	// the diffuse color buffer
	mFbo->AddColorBuffer(ColorBufferObject::BUFFER_UBYTE, ColorBufferObject::WRAP_CLAMP_TO_EDGE, ColorBufferObject::FILTER_LINEAR, false);

	mShadowCam = new Camera(mSceneBox.Size().x * 0.5f, mSceneBox.Size().y * 0.5f);
	mShadowCam->SetOrtho(true);
}


ShadowMapping::~ShadowMapping()
{
	if (sCgShadowProgram)
		cgDestroyProgram(sCgShadowProgram);

	DEL_PTR(mFbo);
}


void ShadowMapping::Init(CGcontext context)
{/*
	sCgShadowProgram = 
		cgCreateProgramFromFile(context, 
								CG_SOURCE,
								"src/shaders/shadow.cg", 
								RenderState::sCgFragmentProfile,
								"main",
								NULL);

	if (sCgShadowProgram != NULL)
	{
		cgGLLoadProgram(sCgShadowProgram);
		sShadowParam = cgGetNamedParameter(sCgShadowProgram, "shadowMap");
	}
	else
		cerr << "shadow program failed to load" << endl;
*/
	PrintGLerror("init");
}


void ShadowMapping::ComputeShadowMap(Light *light, RenderTraverser *traverser)
{
	mLight = light;

	const float xlen = Magnitude(mSceneBox.Diagonal() * 0.5f);
	const float ylen = Magnitude(mSceneBox.Diagonal() * 0.5f);
	
	mShadowCam->SetDirection(light->GetDirection());
	
	// set position so that we can see the whole scene
	Vector3 pos = mSceneBox.Center();
	pos.z = mSceneBox.Max().z;
	//mShadowCam->SetPosition(pos - Magnitude(mSceneBox.Diagonal() * light->GetDirection()));
	mShadowCam->SetPosition(pos);

	mFbo->Bind();

	glDrawBuffers(1, mrt);

	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0, 0, mSize, mSize);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

	glPolygonOffset(1.0f, 4000.0f);
	//glEnable(GL_POLYGON_OFFSET_FILL);

	glShadeModel(GL_FLAT);
	glEnable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();

	glOrtho(+xlen, -xlen, ylen, -ylen, 0.0f, Magnitude(mSceneBox.Diagonal())); 

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	mShadowCam->SetupCameraView();

	mShadowCam->GetModelViewMatrix(mShadowMatrix);

	traverser->RenderScene();
	
	glDisable(GL_POLYGON_OFFSET_FILL);

	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glPopAttrib();
	
	/*float *data = new float[mSize * mSize];

	GrabDepthBuffer(data, mFbo->GetDepthTex());
	ExportDepthBuffer(data, mSize);

	delete [] data;*/

	FrameBufferObject::Release();
}

/*
void ShadowMapping::Render(RenderTraverser *traverser, Camera *cam)
{
	Matrix4x4 cam_proj;
	Matrix4x4 cam_inverse_modelview;

	const float bias[16] = 
	{0.5f, 0.0f, 0.0f, 0.0f, 
	0.0f, 0.5f, 0.0f, 0.0f,
	0.0f, 0.0f, 0.5f, 0.0f,
	0.5f, 0.5f, 0.5f, 1.0f	};


	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// update the camera, so that the user can have a free look
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	cam->SetupCameraView();

	// store the inverse of the resulting modelview matrix for the shadow computation
	glGetFloatv(GL_MODELVIEW_MATRIX, (float *)cam_inverse_modelview.x);
	cam_inverse_modelview.Invert();

	// bind shadow map
	glBindTexture(GL_TEXTURE_2D_ARRAY_EXT, mFbo->GetDepthTex());

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
	//glTexParameteri(GL_TEXTURE_2D_ARRAY_EXT, GL_TEXTURE_COMPARE_MODE, GL_NONE);

	PrintGLerror("texture");

	//float light_m[16];
	
	// compute a matrix that transforms from camera eye space to light clip space
	// and pass it to the shader through the OpenGL texture matrices, since we
	// don't use them now
	glEnable(GL_TEXTURE_2D);
	glMatrixMode(GL_TEXTURE);

	glLoadMatrixf(bias);
	glMultMatrixf((float *)mShadowMatrix.x);
	
	// multiply the light's (bias * crop * proj * modelview) by the inverse camera modelview
	// so that we can transform a pixel as seen from the camera
	glMultMatrixf((float *)cam_inverse_modelview.x);
	
		
	cgGLEnableProfile(RenderState::sCgFragmentProfile);
	cgGLBindProgram(sCgShadowProgram);
	
	cgGLSetTextureParameter(sShadowParam, mFbo->GetDepthTex());
	cgGLEnableTextureParameter(sShadowParam);
	
	// finally, draw the scene 
	traverser->RenderScene();

	cgGLDisableTextureParameter(sShadowParam);

	PrintGLerror("shadow");
}
*/


//Called to draw scene
void ShadowMapping::Render(RenderTraverser *traverser, Camera *cam)
{
	Matrix4x4 cam_inverse_modelview;

	//Calculate texture matrix for projection
	//This matrix takes us from eye space to the light's clip space
	//It is postmultiplied by the inverse of the current view matrix when specifying texgen
	static Matrix4x4 biasMatrix(0.5f, 0.0f, 0.0f, 0.0f,
								0.0f, 0.5f, 0.0f, 0.0f,
								0.0f, 0.0f, 0.5f, 0.0f,
								0.5f, 0.5f, 0.5f, 1.0f); //bias from [-1, 1] to [0, 1]

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// update the camera, so that the user can have a free look
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	cam->SetupCameraView();

	// store the inverse of the resulting modelview matrix for the shadow computation
	glGetFloatv(GL_MODELVIEW_MATRIX, (float *)cam_inverse_modelview.x);
	cam_inverse_modelview.Invert();

	// bind shadow map
	glBindTexture(GL_TEXTURE_2D_ARRAY_EXT, mFbo->GetDepthTex());

	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_NONE);

	PrintGLerror("texture");
	
	// compute a matrix that transforms from camera eye space to light clip space
	// and pass it to the shader through the OpenGL texture matrices, since we
	// don't use them now
	glEnable(GL_TEXTURE_2D);
	glMatrixMode(GL_TEXTURE);

	glLoadMatrixf((float *)biasMatrix.x);
	glMultMatrixf((float *)mShadowMatrix.x);
	
	// multiply the light's (bias * crop * proj * modelview) by the inverse camera modelview
	// so that we can transform a pixel as seen from the camera
	glMultMatrixf((float *)cam_inverse_modelview.x);

	RgbaColor white(1, 1, 1, 0);
	RgbaColor dark(0.2, 0.2, 0.2, 0);

	//Use dim light to represent shadowed areas
	glLightfv(GL_LIGHT1, GL_POSITION, mLight->GetDirection());
	glLightfv(GL_LIGHT1, GL_AMBIENT, (float *)&dark.r);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, (float *)&white.r);
	//glLightfv(GL_LIGHT1, GL_SPECULAR, black);
	glEnable(GL_LIGHT1);
	glEnable(GL_LIGHTING);

	
	// finally, draw the scene 
	traverser->RenderScene();

	

	//3rd pass
	//Draw with bright light
	glLightfv(GL_LIGHT1, GL_DIFFUSE, (float *)&white.r);
	glLightfv(GL_LIGHT1, GL_SPECULAR, (float *)&white.r);

	Matrix4x4 textureMatrix;// = biasMatrix * mLightProjectionMatrix * mLightViewMatrix;

	//Set up texture coordinate generation.
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	//glTexGenfv(GL_S, GL_EYE_PLANE, textureMatrix.GetRow(0));
	glEnable(GL_TEXTURE_GEN_S);

	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	//glTexGenfv(GL_T, GL_EYE_PLANE, textureMatrix.GetRow(1));
	glEnable(GL_TEXTURE_GEN_T);

	glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	//glTexGenfv(GL_R, GL_EYE_PLANE, textureMatrix.GetRow(2));
	glEnable(GL_TEXTURE_GEN_R);

	glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	//glTexGenfv(GL_Q, GL_EYE_PLANE, textureMatrix.GetRow(3));
	glEnable(GL_TEXTURE_GEN_Q);

	//Bind & enable shadow map texture
	glBindTexture(GL_TEXTURE_2D, mFbo->GetDepthTex());
	glEnable(GL_TEXTURE_2D);

	//Enable shadow comparison
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE);

	//Shadow comparison should be true (ie not in shadow) if r<=texture
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);

	//Shadow comparison should generate an INTENSITY result
	glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE_ARB, GL_INTENSITY);

	//Set alpha test to discard false comparisons
	glAlphaFunc(GL_GEQUAL, 0.99f);
	glEnable(GL_ALPHA_TEST);

	traverser->RenderScene();

	//Disable textures and texgen
	glDisable(GL_TEXTURE_2D);

	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_TEXTURE_GEN_R);
	glDisable(GL_TEXTURE_GEN_Q);

	//Restore other states
	glDisable(GL_LIGHTING);
	glDisable(GL_ALPHA_TEST);
}


} // namespace
